
/*********************************************************************
 * Squid 2 ClamAV redirector                                         *
 *********************************************************************/

/*********************************************************************
 *                                                                   *
 * ---------+-------------------------------------+--------+-------- *
 *  Version | Aenderung                           | Datum  | Author  *
 * ---------+-------------------------------------+--------+-------- *
 *   0.1    | Program build                       |11.01.05| me      *
 *          |                                     |        |         *
 * ---------+-------------------------------------+--------+-------- *
 *   0.2    | Added Database reloading            |17.01.05| me      *
 *          | fixed some bugs                     |        |         *
 * ---------+-------------------------------------+--------+-------- *
 *   0.3    | Added some optional debugging output|19.01.05| me      *
 *          | to fix a strange bug                |        |         *
 * ---------+-------------------------------------+--------+-------- *
 *   0.4    | see changelog for further changes   |22.01.05| me      *
 * ---------+-------------------------------------+--------+-------- *  
 *********************************************************************/

/*********************************************************************
 * Copyright (C) 2005 Daniel Lord (squidclam At users DoT sf Dot net)*
 *                                                                   *
 * This is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License as published *
 * by the Free Software Foundation; either version 2 of the License, *
 * or (at your option) any later version.                            *
 *                                                                   *
 * This software is distributed in the hope that it will be useful,  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 * GNU General Public License for more details.                      *
 *                                                                   *
 * You should have received a copy of the GNU General Public License *
 * along with this software; if not, write to the Free Software      *
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,            *
 * MA  02111-1307, USA.                                              *
 *********************************************************************/

/* return codes
 * 0 - Success 
 * 1 - Could not unbuffer stdout
 * 2 - problem with clamav
 * 3 - problem with temp files
 * 4 - problem with curl
 */

/* includes  */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <syslog.h>
#include <clamav.h>
#include <curl/curl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>

/* defines  */
#ifndef ERROR
	#define ERROR    "http://127.0.0.1/antivir.php"    /* url with the antivir.php file */
#endif
#ifndef MY_PROXY
	#define MY_PROXY "http://127.0.0.1:3128"           /* proxy to use */
#endif
#ifndef TMPF
	#define TMPF     "/tmpdata/squidclamXXXXXXXX"      /* tempfile preferable on ramdisk */
#endif
#ifndef RLDB
	#define RLDB  (100)            /* reload db after how many url + random #(1-255) */
#endif
#ifndef FSIZE
	#define FSIZE (204800)         /* scan/get max 100kb of data */
#endif
#ifndef CONFIG
	#define CONFIG "/etc/squidclam.conf"
#endif

/* if defined hand url over to squid and don't do anything in case of error */
#define ERRIGNORE
/* if defined display debug output */
// #define DEBUG

/* no changes needed down there */
#define TRUE  1
#define FALSE 0

/* Memory struct prototype  */
struct MemoryStruct {
      char *memory;
      size_t size;
};

/* config struct prototype */
struct cfg {
    char *proxy;
    char *url;
    char *tmp;
    unsigned int rldb;
    unsigned int fsize;
};

/* open config file */
FILE * 
opencfg (void)
{
    FILE *fd=NULL;
    
    if ((fd = fopen(CONFIG, "rb")) == NULL) {
        fprintf(stderr,
				"failed to open config (%s) using hardcoded values\n"
				, CONFIG);
    }

    return fd;
}

/* close config file */
int 
closecfg (FILE *fd)
{
    fclose(fd);
    return 0;
}

/* strip blanks */
char * 
stripblank (char *p) {
	char *n;
	unsigned int l=0, x=0, i=0;
	
#ifdef DEBUG
	fprintf(stderr,"string befor (%s)\n", p);
#endif

	
	/* get space for the new string */
	l = strlen(p);
	n = malloc( l+1 );
	
	/* check and make clean new string */
	for (i=0; i<l; i++) {
		if ( ( p[i] != ' ' ) && (p[i] != '\t') ) {
			n[x] = p[i];
			x++;
		}
	}
	
	/* make sure string is terminated */
	n[x] = 0x00;

#ifdef DEBUG
	fprintf(stderr,"string after (%s)\n", n);
#endif
	
	return n;
}

/* strip \r\n and some more */
char * 
stripctrl (char *s) {
	char *p;
	
	for (p=s;*p;p++) {
		if (iscntrl(*p) != 0) {
			*p = 0x00;
		}
	}
	
	return s;
}

/* check if this is a valid cfg line (no comment eg.) */
int 
is_cfg_line (char *p) {
#ifdef DEBUG
	char t[] = { " Line recogniced as comment\n" };
#endif
	
	/* if line starts with a comment mark or is empty 
	 * return FALSE */
	if ((*p == '#') || (*p == '\0')) {
#ifdef DEBUG
		fprintf(stderr,"%s%s", p, t);
#endif
		return FALSE;
	}
	/* TODO */
	return TRUE;
}

/* get the value of the config variable */
char * 
getvalue (char *s) {
	char *x;
	
	x = strstr(s,"=");
	
	return ++x;
}

int 
partial_cont (char *s) {
	
	/* check if we read the whole line */
	if (s[strlen(s)-1] != '\n') {
		/* if not complete we have partial content so return TRUE */
		return TRUE;
	}
	
	return FALSE;
}

/* read config and init vars */
struct cfg 
read_cfg (FILE *f)
{
    struct cfg cfg;
	char ts[100];
	char *p;

    cfg.proxy = NULL;
    cfg.url = NULL;
    cfg.tmp = NULL;
    cfg.rldb = 0;
    cfg.fsize = 0;
	
	if (f != NULL) {
		while (((fgets(ts,sizeof(ts),f)) != NULL) && (partial_cont(ts) != TRUE)) {
			p = stripblank(ts);
			p = stripctrl(p);
			if (is_cfg_line(p) == TRUE) {
				switch(*p) {
					case 'p':
						p = getvalue(p);
						cfg.proxy = malloc(strlen(p)+1);
						strcpy(cfg.proxy, p);
						break;
					case 'u':
						p = getvalue(p);
						cfg.url = malloc(strlen(p)+1);
						strcpy(cfg.url, p);
						break;
					case 't':
						p = getvalue(p);
						cfg.tmp = malloc(strlen(p)+1);
						strcpy(cfg.tmp, p);
						break;
					case 'r':
						p = getvalue(p);
						cfg.rldb = atoi(p);
						break;
					case 'f':
						p = getvalue(p);
						cfg.fsize = atoi(p);
						break;
					default:
						fprintf(stderr,
								"wrong config line (%s)\n",ts);
						break;
				}
			}
		}
	}

	if (cfg.proxy == NULL) {
		cfg.proxy = malloc(sizeof(MY_PROXY)+1);
		strncpy(cfg.proxy, MY_PROXY, sizeof(MY_PROXY)); /* no strncpy cause sizeof is used befor */
	}
	if (cfg.url == NULL) {
		cfg.url = malloc(sizeof(ERROR)+1);
		strncpy(cfg.url, ERROR, sizeof(ERROR)); /* no strncpy cause sizeof is used befor */
	}
	if (cfg.tmp == NULL) {
		cfg.tmp = malloc(sizeof(TMPF)+1);
		strncpy(cfg.tmp, TMPF,sizeof(TMPF)); /* no strncpy cause sizeof is used befor */
	}
	if (cfg.rldb <= 0) {
		cfg.rldb = RLDB;
	}
	if (cfg.fsize <= 0) {
		cfg.fsize = FSIZE;
	}

    return cfg;
}

/* Memory funktion to store the download  */
size_t
WriteMemoryCallback(void *ptr, size_t size, size_t nmemb, void *data)
{
      register int realsize = size * nmemb;
      struct MemoryStruct *mem = (struct MemoryStruct *)data;
      
      mem->memory = (char *)realloc(mem->memory, mem->size + realsize + 1);
      
      if (mem->memory) {
          memcpy(&(mem->memory[mem->size]), ptr, realsize);
          mem->size += realsize;
          mem->memory[mem->size] = 0;
      }
      
      return realsize;
}

/* get random number  */
unsigned int 
get_random_number (void)
{
    int rf;
    unsigned char ret[1];

    /* open random source */
    if ((rf = open("/dev/urandom", O_RDONLY)) == -1) {
        usleep(1000000);
        return -1;
    }
    
    /* get one random byte */
    if (read(rf,ret,1) != 1) {
        return -1;
    }
    
    /* close random source */
    close(rf);
    
    return ret[0];
}

/* one bloated main for the moment  */
int 
main (void)
{
    const char *dbdir;
    const char *virname;
    struct cl_node *root = NULL;
    struct cl_limits limits;

	struct cfg cfg;

    struct request {
        char url[10000]; /* don't forget to also change scanf if you tweak this */
        char src[255];   /* don't forget to also change scanf if you tweak this */
        char ident[255]; /* don't forget to also change scanf if you tweak this */
        char method[31]; /* don't forget to also change scanf if you tweak this */
    } rq;

    CURL *eh;
    double dsize=0;
    
    struct MemoryStruct mem;

    int  ret=0;
    unsigned long int size = 0;
    int  fd;
	char tt[255]; /* please give me a hint how to avoid this crap... */
    char erbuf[CURL_ERROR_SIZE+1];
    
    unsigned int count=0;
    unsigned int rand=0;
    unsigned int reload=0;

	FILE *cf; /* config file handle */

    /* init struct */
    mem.memory=NULL; /* we expect realloc(NULL, size) to work */
    mem.size = 0;    /* no data at this point */

   
	
    /* get random number */
    if ((rand = get_random_number()) == -1) {
        syslog(LOG_ERR, "random number failed");
        rand=10; /* default value of rand */
    }

    /* start syslogging */
    openlog("squidclam", LOG_PID, LOG_DAEMON);
    
    /* now sleep a bit to distribute the load at startup */ 
    usleep((rand%10)*10000);

	/* open config */
	cf = opencfg();
	
	/* read config and fill the struct with values */
	cfg = read_cfg(cf);

#ifdef DEBUG
	fprintf(stderr, "\nproxy=%s\nurl=%s\ntmp=%s\nrldb=%i\nfsize=%i\n",
			cfg.proxy, cfg.url, cfg.tmp, cfg.rldb, cfg.fsize);
#endif

	/* close config */
	if (cf != NULL) {
		closecfg(cf);
	}

    /* initialice reload (RLDB + (0-255)) */
    reload = cfg.rldb + rand;

    /* info that we go up */
    syslog(LOG_INFO,"squidclam starting up now. reload after %i URLs", reload);

    /* make stdout line buffered */
    if (setvbuf(stdout, NULL, _IOLBF, 0) != 0) {
        syslog(LOG_ERR, "stdout not line buffered exiting");
        usleep(1000000); /* don't do DoS */
        return 1;
    }

    /* clamav init (load database) */
    dbdir = cl_retdbdir();
    if ((ret=cl_loaddbdir(dbdir, &root, NULL))) {
        syslog(LOG_ERR, "libclamav: %s", cl_strerror(ret));
        usleep(1000000); /* don't do DoS */
        return 2;
    }
    
    /* build database */
    if((ret = cl_build(root))) {
        syslog(LOG_ERR,"libclamav database initialization error: %s", cl_strerror(ret));
        usleep(1000000); /* don't do DoS */
        return 2;
    }

    /* set up archive limits */
    memset(&limits, 0, sizeof(struct cl_limits));
    limits.maxfiles = 20;             /* max files */
    limits.maxfilesize = 2097152;     /* maximal archived file size == 2 Mb */
    limits.maxreclevel = 5;           /* maximal recursion level */
    limits.maxratio = 100;            /* maximal compression ratio */
    limits.archivememlim = 0;         /* unlimited memory for bzip2 scanner */
    
    /* curl init */
    curl_global_init(CURL_GLOBAL_ALL);

    /* get an easy handle */
    if ((eh = curl_easy_init()) == NULL) {
        syslog(LOG_ERR,"curl_easy_init failed\n");
        usleep(1000000); /* don't do DoS */
        return 4;
    }

    /* set the proxy */
    curl_easy_setopt(eh, CURLOPT_PROXY, cfg.proxy);

    /* set http version to fool buggy servers */
    curl_easy_setopt(eh, CURL_HTTP_VERSION_1_0, TRUE);

	/* fail on errors to avoid pebcak problems */
	curl_easy_setopt(eh, CURLOPT_FAILONERROR, TRUE);

	/* set useragent */
	curl_easy_setopt(eh, CURLOPT_USERAGENT, "squidclam");

    /* follow location (303 MOVED) */
    curl_easy_setopt(eh, CURLOPT_FOLLOWLOCATION, TRUE);
    
    /* only follow 5 redirects */
    curl_easy_setopt(eh, CURLOPT_MAXREDIRS, 5);

    /* send all data to this function  */
    curl_easy_setopt(eh, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);

    /* we pass our 'chunk' struct to the callback function */
    curl_easy_setopt(eh, CURLOPT_WRITEDATA, (void *)&mem);

    /* get usefull error messages */
    curl_easy_setopt(eh, CURLOPT_ERRORBUFFER, erbuf);

	/* get a object for mkstemp (bad programming) */ 
	strncpy(tt,cfg.tmp,254);
	tt[255] = 0x00;
   
    /* get a filehandle preferable in a ram disk */
    if ((fd = mkstemp(tt)) == -1) {
        syslog(LOG_ERR,"Could not get tempfile handle for template %s. Read `man mkstemp`", cfg.tmp);
        usleep(1000000); /* don't do DoS */
        return 3;
    }

    /* loop to parse the given URLs */
    while (scanf("%10000s %255s %255s %31s", rq.url, rq.src, rq.ident, rq.method) != EOF){
        /* show which url we're handling */
#ifdef DEBUG
        syslog(LOG_INFO,"handle url (%s) src=(%s) ident=(%s) method=(%s)\n", rq.url, rq.src, rq.ident, rq.method);
#endif
        /* quick "fix" for the POST Problem */
/*        if (rq.method[0] != 'G') {
            syslog(LOG_INFO,"POST request ignored for now (%s)\n", rq.url);
			putchar('\n');
            continue;
        }
*/       
        /* clean struct */
        free(mem.memory);
        mem.memory = NULL;
        mem.size = 0;    /* no data at this point */
    
        /* set the url to retrive */
        curl_easy_setopt(eh, CURLOPT_URL, rq.url);

        /* only get the head */
        curl_easy_setopt(eh, CURLOPT_NOBODY, TRUE);
        
        /* actually get the header */
        if (curl_easy_perform(eh) != 0) {
            /* no header so just go back to squid */
#ifdef ERRIGNORE
			putchar('\n');
#else
            fprintf(stdout,"%s?url=%s&virus=%s %s %s %s\n", cfg.url, rq.url, 
                    "squidclam_get_head_failed", rq.src, rq.ident, rq.method);
#endif
#ifdef DEBUG
            syslog(LOG_INFO,"could not get HEAD (%s) error was (%s)\n", rq.url, erbuf);
#endif
            continue;
        }
        
        /* get the CONTENT_LENGTH out of the header */
        if (curl_easy_getinfo(eh, CURLINFO_CONTENT_LENGTH_DOWNLOAD, &dsize) != CURLE_OK) {
            /* not ok --> squid should handle this */
#ifdef ERRIGNORE
			putchar('\n');
#else
            fprintf(stdout,"%s?url=%s&virus=%s %s %s %s\n", cfg.url, rq.url, 
                    "squidclam_get_size_of_file_failed", rq.src, rq.ident, rq.method);
#endif
#ifdef DEBUG
            syslog(LOG_INFO,"could not get size from (%s) error was (%s)\n", rq.url, erbuf);
#endif
            continue;
        }

        /* file too big --> don't scan */
        if (dsize > FSIZE) {
			putchar('\n');
#ifdef DEBUG
            syslog(LOG_INFO,"file too big to scan. %s has %.0f bytes\n", rq.url, dsize);
#endif
            continue;
        }
        
        /* now get the complete file */
        curl_easy_setopt(eh, CURLOPT_NOBODY, FALSE);
        curl_easy_setopt(eh, CURLOPT_HTTPGET, TRUE);

        /* get the file */
        if (curl_easy_perform(eh) != 0) {
            /* if error, give back to squid */
#ifdef ERRIGNORE
			putchar('\n');
#else
            fprintf(stdout,"%s?url=%s&virus=%s %s %s %s\n", cfg.url, rq.url, 
                    "squidclam_get_file_failed", rq.src, rq.ident, rq.method);
#endif
#ifdef DEBUG
            syslog(LOG_INFO,"could not get (%s) error was (%s)\n", rq.url, erbuf);
#endif
            continue;
        }
        
        /* reset file handle to the beginning */
        lseek(fd,SEEK_SET,SEEK_SET);

        /* get a file out of the memory chunk */
        if (write(fd,mem.memory, mem.size) != mem.size) {
#ifdef ERRIGNORE
			putchar('\n');
#else
            fprintf(stdout,"%s?url=%s&virus=%s %s %s %s\n", cfg.url, rq.url, 
                    "squidclam_writing_tempfile_failed", rq.src, rq.ident, rq.method);
#endif
            syslog(LOG_INFO,"error writing data to tempfile");
            continue;
        }

        /* make sure, there is no extra data in tempfile */
        if (ftruncate(fd, mem.size) != 0) {
#ifdef ERRIGNORE
			putchar('\n');
#else
            fprintf(stdout,"%s?url=%s&virus=%s %s %s %s\n", cfg.url, rq.url, 
                    "squidclam_truncate_tempfile_failed", rq.src, rq.ident, rq.method);
#endif
            syslog(LOG_INFO,"error truncating tempfile");
            continue;
        }

        /* scan the file with clamav */
        /* we sadly have to use files for clamav to work properly */
        /* but we can put them into a ramdisk or tmpfs ;) */
        /* then we only have to deal with insecure tmpfile creation */
        if((ret = cl_scandesc(fd, &virname, &size, root, &limits, CL_SCAN_STDOPT)) == CL_VIRUS) {
            fprintf(stdout,"%s?url=%s&virus=%s %s %s %s\n", cfg.url, rq.url, virname, rq.src, rq.ident, rq.method);
            syslog(LOG_WARNING, "INFECTED url=%s virus=%s src=%s ident=%s method=%s", rq.url, virname, rq.src, rq.ident, rq.method);
#ifdef DEBUG
            syslog(LOG_WARNING, "Size of file: %i\n", mem.size);
#endif
        }
        else {
            /* file seems to be clean */
			putchar('\n');
#ifdef DEBUG
            syslog(LOG_INFO,"Scanned but no Virus found at: (%s)\n", rq.url);
            syslog(LOG_INFO,"Size of file: %i\n", mem.size);
#endif
            if(ret != CL_CLEAN) {
                syslog(LOG_ERR, "%s\n", cl_perror(ret));
            }
        }
        
        if (count++ >= reload) {
#ifdef DEBUG
            syslog(LOG_INFO,"Reloading DB by restarting squidclam at %i's URL", reload);
#endif
            closelog();
            free(mem.memory);
            curl_global_cleanup();
            cl_free(root);
            unlink(cfg.tmp);
            return 0;
            /* update signature db this is done easyly be reloading squidclam :) */
        }
    }
    
    /* clean up  */
    closelog();
    free(mem.memory);
    curl_global_cleanup();
    cl_free(root);
    unlink(cfg.tmp);
    return 0;
}
